import json
from pathlib import Path
from typing import Iterable

from pygit2 import Commit

from girokmoji.catgitmoji import by_gitmoji, any_to_catmoji
from girokmoji.const import CATEGORY, category_order, CATEGORY_SUBTEXTS
from girokmoji.exception import (
    NoGitmojiInMessageError,
    MessageDoesNotStartWithGitmojiError,
    NoSuchGitmojiSupportedError,
)
from girokmoji.git import get_tag_to_tag_commits
from girokmoji.template import ENTRY_GROUP_HEADER, ENTRY_SUBITEM
from girokmoji.template import SEPARATOR, HEAD, CATEGORY_SECTION


def commit_message(commit: Commit) -> str:
    if isinstance(commit.message, str):
        return commit.message

    return commit.raw_message.decode(commit.message_encoding)


def get_category(msg: str, *, fallback_to_includes: bool = True) -> CATEGORY:
    for gitmoji in by_gitmoji():
        if msg.startswith(gitmoji):
            return by_gitmoji()[gitmoji].category
    if fallback_to_includes:
        for gitmoji in by_gitmoji():
            if gitmoji in msg:
                return by_gitmoji()[gitmoji].category
    raise NoGitmojiInMessageError("No Gitmoji found in the message")


def sep_gitmoji_msg_title(msg: str, *, strict: bool = False) -> tuple[str, str]:
    """Return gitmoji and message from commit message. Strict mode raises exception MessageDoesNotStartWithGitmojiError"""
    msg = msg.split("\n")[0]
    for gitmoji in by_gitmoji().keys():
        if msg.startswith(gitmoji):
            return gitmoji, msg.removeprefix(gitmoji).strip()

    if not strict:
        return "", msg.split("\n")[0]

    raise MessageDoesNotStartWithGitmojiError


def structured_changelog(commits: Iterable[Commit]) -> dict[CATEGORY, list[Commit]]:
    # prepare structured changelog with importance order
    structured_changelog: dict[CATEGORY, list[Commit]] = {}
    for cat in category_order:
        structured_changelog[cat] = []

    for commit in commits:
        msg = commit_message(commit)
        try:
            cat: CATEGORY = get_category(msg)
        except NoGitmojiInMessageError:
            cat = "Hmm..."
        structured_changelog[cat].append(commit)

    return structured_changelog


def gen_markdown(
    project_name: str,
    version: str,
    release_date: str,
    change: dict[CATEGORY, list[Commit]],
):
    changelog_markdown = ""

    separator = SEPARATOR().markdown
    head_md = HEAD(
        project_name=project_name,
        version=version,
        subtext="""
_"Change is always thrilling!"_  
_(And sometimes a little confusing.)_
    """,
        release_date=release_date,
    ).markdown

    changelog_markdown += head_md
    changelog_markdown += separator

    for cat in change:
        category_md = ""
        subcats: dict[tuple[str, str], list[tuple[str, str]]] = {}
        for commit in change[cat]:
            gitmoji, title = sep_gitmoji_msg_title(commit_message(commit))
            if not gitmoji:
                # Ignore commits without a recognizable gitmoji
                continue
            try:
                catmoji = any_to_catmoji(gitmoji)
            except NoSuchGitmojiSupportedError:
                # Skip unsupported gitmoji rather than error out
                continue
            key = (catmoji.emoji, catmoji.description)
            subcats.setdefault(key, []).append((title, str(commit.id)))

        for (emoji, description), items in subcats.items():
            header = ENTRY_GROUP_HEADER(
                emoji=emoji,
                gitmoji_description=description,
            ).markdown
            category_md += header
            for title, commit_hash in items:
                item = ENTRY_SUBITEM(
                    commit_description=title,
                    commit_hash=commit_hash,
                ).markdown
                category_md += item
        if category_md:
            changelog_markdown += CATEGORY_SECTION(cat, CATEGORY_SUBTEXTS[cat]).markdown
            changelog_markdown += category_md
            changelog_markdown += separator

    return changelog_markdown


def change_log(
    project_name: str,
    release_date: str,
    repo_dir: Path,
    tail_tag: str,
    head_tag: str,
    version: str | None = None,
) -> str:
    if version is None:
        version = head_tag

    return gen_markdown(
        project_name,
        version,
        release_date,
        structured_changelog(get_tag_to_tag_commits(repo_dir, tail_tag, head_tag)),
    ).strip()


def github_release_payload(
    project_name: str,
    release_date: str,
    repo_dir: Path,
    tail_tag: str,
    head_tag: str,
    version: str | None = None,
    *,
    draft: bool = False,
    prerelease: bool = False,
) -> str:
    """Return GitHub release payload as JSON string."""
    changelog = change_log(
        project_name=project_name,
        release_date=release_date,
        repo_dir=repo_dir,
        tail_tag=tail_tag,
        head_tag=head_tag,
        version=version,
    )
    payload = {
        "tag_name": head_tag,
        "name": version or head_tag,
        "body": changelog,
        "draft": draft,
        "prerelease": prerelease,
    }
    return json.dumps(payload)



from pathlib import Path
from typing import Iterable

from pygit2 import Commit, Repository, GIT_SORT_TOPOLOGICAL, discover_repository
from pygit2.enums import ObjectType

from girokmoji.exception import NoSuchTagFoundError


def get_tag_to_tag_commits(
    repo_dir: Path, tail_tag: str, head_tag: str
) -> Iterable[Commit]:
    repo = Repository(discover_repository(str(repo_dir)))
    head_ref = repo.references.get(f"refs/tags/{head_tag}")
    tail_ref = repo.references.get(f"refs/tags/{tail_tag}")
    if head_ref is None:
        raise NoSuchTagFoundError(f"{head_tag} can't be found")
    if tail_ref is None:
        raise NoSuchTagFoundError(f"{tail_tag} can't be found")

    head_commit = repo[head_ref.target].peel(ObjectType.COMMIT)
    tail_commit = repo[tail_ref.target].peel(ObjectType.COMMIT)

    rev_walk = repo.walk(head_commit.id, GIT_SORT_TOPOLOGICAL)
    rev_walk.hide(tail_commit.id)
    for rev in rev_walk:
        if isinstance(rev, Commit):
            yield rev



from typing import Literal

CATEGORY = Literal[
    "Critical Changes",
    "Feature and Functional Changes",
    "Bug Fixes",
    "Performance Improvements",
    "Lint",
    "Documentation and Comment",
    "Test",
    "Code Maintenance and Refactoring",
    "Dependency, Build, and Configuration",
    "File and Project Management",
    "Internalization, Accessibility, and UI/UX",
    "Miscellaneous / Other Changes",
    "Hmm...",
]
category_order: list[CATEGORY] = [
    "Critical Changes",
    "Feature and Functional Changes",
    "Bug Fixes",
    "Performance Improvements",
    "Lint",
    "Documentation and Comment",
    "Test",
    "Code Maintenance and Refactoring",
    "Dependency, Build, and Configuration",
    "File and Project Management",
    "Internalization, Accessibility, and UI/UX",
    "Miscellaneous / Other Changes",
    "Hmm...",
]
CATEGORY_SUBTEXTS: dict[CATEGORY, str] = {
    "Critical Changes": "*These are the changes that might keep you up at night.*",
    "Feature and Functional Changes": '*New features that make you go "Wow!"*',
    "Bug Fixes": "*Stubborn bugs have been squashed!*",
    "Performance Improvements": "*Our code is now as energetic as a fresh cup of coffee!*",
    "Lint": "*Polishing code style and eliminating lint warnings!*",
    "Documentation and Comment": "*Documentation and comments are the developer’s conscience.*",
    "Test": "*Tests provide our code with the stability it deserves.*",
    "Code Maintenance and Refactoring": "*Time to tidy up the code and say goodbye to technical debt.*",
    "Dependency, Build, and Configuration": "*From dependency management to build scripts, everything has been updated!*",
    "File and Project Management": "*Managing file moves, removals, and project documentation.*",
    "Internalization, Accessibility, and UI/UX": "*Making the project accessible and appealing to everyone worldwide!*",
    "Miscellaneous / Other Changes": "*A catch-all for changes that don’t quite fit elsewhere.*",
    "Hmm...": "*These changes might seem odd, but somehow they’re endearing.*",
}
SEMVER = Literal["major", "minor", "patch", None]



import json
from pathlib import Path

import pytest

from girokmoji import catgitmoji
from girokmoji import changelog


class FakeCommit:
    def __init__(self, message: str, commit_id: str = "deadbeef"):
        self.message = message
        self.raw_message = message.encode()
        self.message_encoding = "utf-8"
        self.id = commit_id


def test_commit_message_str_bytes():
    commit = FakeCommit(":sparkles: feat")
    assert changelog.commit_message(commit) == ":sparkles: feat"
    commit.raw_message = b":bug: fix"
    commit.message = None
    assert changelog.commit_message(commit) == ":bug: fix"


def test_get_category_and_sep_title():
    msg = ":art: refactor"
    assert changelog.get_category(msg) == catgitmoji.by_code()[":art:"].category
    emoji, title = changelog.sep_gitmoji_msg_title(msg, strict=True)
    assert emoji == ":art:"
    assert title == "refactor"
    assert changelog.sep_gitmoji_msg_title("no gitmoji", strict=False) == (
        "",
        "no gitmoji",
    )
    with pytest.raises(changelog.MessageDoesNotStartWithGitmojiError):
        changelog.sep_gitmoji_msg_title("no gitmoji", strict=True)


def test_structured_and_markdown(monkeypatch):
    commits = [FakeCommit(":sparkles: first"), FakeCommit(":sparkles: second")]
    structured = changelog.structured_changelog(commits)
    assert structured[catgitmoji.by_code()[":sparkles:"].category][0] is commits[0]

    def fake_get_tag_to_tag_commits(repo_dir, tail_tag, head_tag):
        return commits

    monkeypatch.setattr(
        changelog, "get_tag_to_tag_commits", fake_get_tag_to_tag_commits
    )

    md = changelog.change_log("proj", "2024-01-01", Path("."), "v0", "v1")
    assert md.count("Introduce new features.") == 1
    assert "first" in md and "second" in md
    payload = json.loads(
        changelog.github_release_payload("proj", "2024-01-01", Path("."), "v0", "v1")
    )
    assert payload["tag_name"] == "v1"



import pytest

from girokmoji import catgitmoji


def test_lookup_functions():
    first = catgitmoji.RAW[0]
    assert catgitmoji.by_code()[first.code] is first
    assert catgitmoji.by_entity()[first.entity] is first
    assert catgitmoji.by_emoji()[first.emoji] is first
    assert catgitmoji.by_gitmoji()[first.code] is first
    assert catgitmoji.any_to_catmoji(first.code) is first
    with pytest.raises(catgitmoji.NoSuchGitmojiSupportedError):
        catgitmoji.any_to_catmoji(":unknown:")



from pathlib import Path

import pytest
from pygit2 import init_repository, Signature
from pygit2.enums import ObjectType

from girokmoji import changelog, catgitmoji
from girokmoji.exception import NoGitmojiInMessageError, NoSuchTagFoundError
from girokmoji.git import get_tag_to_tag_commits
from girokmoji.template import SupportTemplate


class FakeCommit:
    def __init__(self, message: str, commit_id: str = "deadbeef"):
        self.message = message
        self.raw_message = message.encode()
        self.message_encoding = "utf-8"
        self.id = commit_id


def test_get_category_fallback_and_error():
    msg = f"prefix {catgitmoji.RAW[0].code} something"
    assert (
        changelog.get_category(msg)
        == catgitmoji.by_code()[catgitmoji.RAW[0].code].category
    )
    with pytest.raises(NoGitmojiInMessageError):
        changelog.get_category("no gitmoji here")


def test_structured_changelog_hmm_category():
    commit = FakeCommit("no gitmoji here")
    res = changelog.structured_changelog([commit])
    assert commit in res["Hmm..."]


def test_get_tag_to_tag_commits_success_and_error(tmp_path):
    repo = init_repository(tmp_path)
    person = Signature("t", "t@example.com")
    f = Path(tmp_path) / "f.txt"
    f.write_text("a")
    repo.index.add_all()
    commit1 = repo.create_commit(
        "HEAD",
        person,
        person,
        ":tada: init",
        repo.index.write_tree(),
        [],
    )
    repo.create_tag("v1", commit1, ObjectType.COMMIT, person, "t1")
    f.write_text("b")
    repo.index.add_all()
    commit2 = repo.create_commit(
        "HEAD",
        person,
        person,
        ":art: change",
        repo.index.write_tree(),
        [commit1],
    )
    repo.create_tag("v2", commit2, ObjectType.COMMIT, person, "t2")
    commits = list(get_tag_to_tag_commits(tmp_path, "v1", "v2"))
    assert [c.id for c in commits] == [commit2]
    with pytest.raises(NoSuchTagFoundError):
        list(get_tag_to_tag_commits(tmp_path, "v1", "v3"))
    with pytest.raises(NoSuchTagFoundError):
        list(get_tag_to_tag_commits(tmp_path, "v9", "v2"))


def test_support_template_markdown_not_implemented():
    with pytest.raises(NotImplementedError):
        _ = SupportTemplate().markdown



from girokmoji import template


def test_templates():
    head = template.HEAD(
        project_name="p",
        version="1",
        subtext="s",
        release_date="today",
    )
    assert "p" in head.markdown
    section = template.CATEGORY_SECTION("Bug Fixes", "sub")
    assert "Bug Fixes" in section.markdown
    entry = template.ENTRY(
        emoji="🐛",
        gitmoji_description="bug",
        commit_description="fix",
        commit_hash="abc",
    )
    assert "abc" in entry.markdown
    header = template.ENTRY_GROUP_HEADER(
        emoji="🐛",
        gitmoji_description="bug",
    )
    assert "bug" in header.markdown
    sub = template.ENTRY_SUBITEM(commit_description="fix", commit_hash="abc")
    assert "fix" in sub.markdown
    sep = template.SEPARATOR()
    assert "---" in sep.markdown



import argparse
import sys
from pathlib import Path

from girokmoji.changelog import change_log, github_release_payload


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Generate a changelog from gitmoji-based commits between two tags."
    )
    parser.add_argument("project_name", help="Name of the project")
    parser.add_argument("release_date", help="Release date (YYYY-MM-DD)")
    parser.add_argument("repo_dir", type=Path, help="Path to the git repository")
    parser.add_argument("tail_tag", help="Older git tag (tail tag)")
    parser.add_argument("head_tag", help="Newer git tag (head tag)")
    parser.add_argument(
        "--release-version",
        dest="version",
        help="Optional release version string (defaults to head_tag)",
        default=None,
    )
    parser.add_argument(
        "--github-payload",
        action="store_true",
        help="Output GitHub Release payload JSON instead of markdown",
    )

    args = parser.parse_args()

    if args.github_payload:
        payload = github_release_payload(
            project_name=args.project_name,
            release_date=args.release_date,
            repo_dir=args.repo_dir,
            tail_tag=args.tail_tag,
            head_tag=args.head_tag,
            version=args.version,
        )
        print(payload, file=sys.stdout)
    else:
        changelog = change_log(
            project_name=args.project_name,
            release_date=args.release_date,
            repo_dir=args.repo_dir,
            tail_tag=args.tail_tag,
            head_tag=args.head_tag,
            version=args.version,
        )
        print(changelog, file=sys.stdout)


if __name__ == "__main__":
    try:
        main()
    except Exception as exc:  # pragma: no cover - top level entry
        print(exc, file=sys.stderr)
        sys.exit(1)



from dataclasses import dataclass
from string import Template
from typing import ClassVar

from girokmoji.const import CATEGORY


class SupportTemplate:
    markdown_template: ClassVar[str]

    @property
    def markdown(self):
        raise NotImplementedError


@dataclass
class Head(SupportTemplate):
    project_name: str
    version: str
    subtext: str
    release_date: str


@dataclass
class CategorySection(SupportTemplate):
    category: CATEGORY
    subtext: str


@dataclass
class Entry(SupportTemplate):
    emoji: str
    gitmoji_description: str
    commit_description: str
    commit_hash: str


@dataclass
class Separator(SupportTemplate):
    pass


class DefaultHead(Head):
    markdown_template = """
# 🚀 **$project_name** Release Changelog $version

$subtext

**Release Date:** $release_date

"""

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute(
            project_name=self.project_name,
            version=self.version,
            subtext=self.subtext,
            release_date=self.release_date,
        )


class DefaultCategorySection(CategorySection):
    markdown_template: ClassVar[str] = """
## $category

*$subtext*

"""

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute(
            category=self.category,
            subtext=self.subtext,
        )


class DefaultEntry(Entry):
    markdown_template: ClassVar[str] = (
        "- **$emoji $gitmoji_description**: [*$commit_description*](../../commit/$commit_hash)\n"
    )

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute(
            emoji=self.emoji,
            gitmoji_description=self.gitmoji_description,
            commit_description=self.commit_description,
            commit_hash=self.commit_hash,
        )


@dataclass
class EntryGroupHeader(SupportTemplate):
    emoji: str
    gitmoji_description: str


class DefaultEntryGroupHeader(EntryGroupHeader):
    markdown_template: ClassVar[str] = "- **$emoji $gitmoji_description**\n"

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute(
            emoji=self.emoji,
            gitmoji_description=self.gitmoji_description,
        )


@dataclass
class EntrySubItem(SupportTemplate):
    commit_description: str
    commit_hash: str


class DefaultEntrySubItem(EntrySubItem):
    markdown_template: ClassVar[str] = (
        "  - [*$commit_description*](../../commit/$commit_hash)\n"
    )

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute(
            commit_description=self.commit_description,
            commit_hash=self.commit_hash,
        )


class DefaultSeparator(Separator):
    markdown_template: ClassVar[str] = """
---

"""

    @property
    def markdown(self):
        return Template(self.markdown_template).substitute()


HEAD = DefaultHead
CATEGORY_SECTION = DefaultCategorySection
ENTRY = DefaultEntry
ENTRY_GROUP_HEADER = DefaultEntryGroupHeader
ENTRY_SUBITEM = DefaultEntrySubItem
SEPARATOR = DefaultSeparator
